SystemOrganization addCategory: #'Seaside-Pharo-Continuation'!

GRObject subclass: #WAContinuation
	instanceVariableNames: 'values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Pharo-Continuation'!

!WAContinuation commentStamp: 'pmm 6/1/2008 00:59' prior: 0!
A WAContinuation is the Squeak implementation of a continuation. It serializes the whole stack into an array

Instance Variables
	values:		<Array>

values
	- an Array of ContextParts and the values they store
!

!WAContinuation class methodsFor: 'instance creation' stamp: 'ab 6/15/2003 19:13'!
current
	^ self fromContext: thisContext sender! !

!WAContinuation class methodsFor: 'instance creation' stamp: 'ab 6/15/2003 19:13'!
currentDo: aBlock
	^ aBlock value: (self fromContext: thisContext sender)! !

!WAContinuation class methodsFor: 'instance creation' stamp: 'jf 3/13/2009 18:05'!
fromContext: aStack
	^ self basicNew initializeFromContext: aStack! !

!WAContinuation methodsFor: 'private' stamp: 'jf 3/13/2009 18:06'!
initializeFromContext: aContext
	| context stream |
	self initialize.
	context := aContext.
	stream := WriteStream on: (Array new: 200).
	[ context notNil ] whileTrue: [
		stream nextPut: context.
		1 to: context class instSize do: [ :index |
			stream nextPut: (context instVarAt: index) ].
		1 to: context size do: [ :index |
			stream nextPut: (context at: index) ].
		context := context sender ].
	values := stream contents! !

!WAContinuation methodsFor: 'accessing' stamp: 'ab 6/15/2003 19:18'!
numArgs
	^ 1! !

!WAContinuation methodsFor: 'private' stamp: 'lr 1/24/2009 21:20'!
restoreValues
	| stream context |
	stream := values readStream.
	[ stream atEnd ] whileFalse: [
		context := stream next.
		1 to: context class instSize do: [ :index |
			context instVarAt: index put: stream next ].
		1 to: context size do: [ :index |
			context at: index put: stream next ] ]! !

!WAContinuation methodsFor: 'private' stamp: 'lr 1/24/2009 21:22'!
terminate: aContext
	| context |
	context := aContext.
	[ context notNil ] 
		whileTrue: [ context := context swapSender: nil ]! !

!WAContinuation methodsFor: 'evaluating' stamp: 'ab 6/15/2003 19:13'!
value
	self value: nil! !

!WAContinuation methodsFor: 'evaluating' stamp: 'lr 10/28/2007 14:42'!
value: anObject
	"Invoke the continuation and answer anObject as return value."

	self terminate: thisContext.
	self restoreValues.
	thisContext swapSender: values first.
	^ anObject! !

!WAContinuation methodsFor: 'evaluating' stamp: 'jf 9/30/2009 00:24'!
valueWithArguments: anArray 
	anArray size = 1 ifFalse: 
		[ ^ GRInvalidArgumentCount signal: 'Continuations can only be resumed with one argument.' ].
	self value: anArray first! !

!WAContinuation methodsFor: 'evaluating' stamp: 'pmm 1/10/2009 23:00'!
valueWithPossibleArguments: anArray
	^ self value: anArray first! !

GRObject subclass: #WAPartialContinuation
	instanceVariableNames: 'values'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Pharo-Continuation'!

!WAPartialContinuation class methodsFor: 'instance creation' stamp: 'jf 2/7/2009 14:04'!
currentDo: aBlock
	| marker |
	marker := WAMarkerRequest signal
		ifNil: [ WAMarkerNotFound signal: 'Marker not found when capturing partial continuation.' ].
	^ aBlock value: (self from: thisContext sender to: marker)! !

!WAPartialContinuation class methodsFor: 'instance creation' stamp: 'lr 2/5/2009 08:47'!
from: aSourceContext to: aTargetContext
	^ self basicNew initializeFrom: aSourceContext to: aTargetContext! !

!WAPartialContinuation methodsFor: 'private' stamp: 'jf 2/6/2009 21:13'!
captureFrom: aSourceContext to: aTargetContext
	"Utilitiy method to capture the contexts between aSourceContext and aTargetContext."

	| stream context |
	context := aSourceContext.
	stream := WriteStream on: (Array new: 100).
	[ context isNil or: [ context == aTargetContext ] ] whileFalse: [
		stream nextPut: context.
		1 to: context class instSize do: [ :index |
			stream nextPut: (context instVarAt: index) ].
		1 to: context size do: [ :index |
			stream nextPut: (context at: index) ].
		context := context sender ].
	context isNil
		ifTrue: [ WAMarkerNotFound signal: aTargetContext printString , ' not found while capturing continuation.' ].
	^ stream contents! !

!WAPartialContinuation methodsFor: 'initialization' stamp: 'lr 2/5/2009 08:48'!
initializeFrom: aSourceContext to: aTargetContext
	self initialize.
	values := self captureFrom: aSourceContext to: aTargetContext! !

!WAPartialContinuation methodsFor: 'accessing' stamp: 'lr 1/14/2009 21:18'!
numArgs
	^ 1! !

!WAPartialContinuation methodsFor: 'private' stamp: 'lr 2/5/2009 08:58'!
restoreValues
	"Restore the captured context chain."

	| stream context |
	stream := values readStream.
	[ stream atEnd ] whileFalse: [ 
		context := stream next.
		1 to: context class instSize do: [ :index | 
			context instVarAt: index put: stream next ].
		1 to: context size do: [ :index | 
			context at: index put: stream next ] ].
	^ context! !

!WAPartialContinuation methodsFor: 'private' stamp: 'lr 2/5/2009 08:54'!
unwindFrom: aSourceContext to: aTargetContext
	"Utilitiy method to unwind and terminate from aSourceContext to aTargetContext. This implementation uses Squeak specific code implemented as primitives and is pretty fast."

	aSourceContext unwindTo: aTargetContext.
	aSourceContext terminateTo: aTargetContext! !

!WAPartialContinuation methodsFor: 'evaluating' stamp: 'lr 1/16/2009 23:16'!
value
	^ self value: nil! !

!WAPartialContinuation methodsFor: 'evaluating' stamp: 'jf 2/6/2009 21:28'!
value: anObject
	"Evaluate the receiving partial continuation. Note tha partial continuations can be evaluated without a marker, in this case they are just spliced at the end of the current stack and resume return back to the sender."

	| marker |
	marker := WAMarkerRequest signal.
	marker isNil
		ifTrue: [ marker := thisContext sender ]
		ifFalse: [ self unwindFrom: thisContext to: marker ].
	self restoreValues swapSender: marker.
	thisContext swapSender: values first.
	^ anObject! !

!WAPartialContinuation methodsFor: 'evaluating' stamp: 'jf 9/30/2009 00:24'!
valueWithArguments: anArray 
	^ anArray size = 1 
		ifTrue: [ self value: anArray first ]
		ifFalse: 
			[ GRInvalidArgumentCount signal: 'Continuations can only be resumed with one argument.' ]! !

!WAPartialContinuation methodsFor: 'evaluating' stamp: 'lr 1/14/2009 21:18'!
valueWithPossibleArguments: anArray
	^ self value: anArray first! !

!GRPackage class methodsFor: '*seaside-pharo-continuation' stamp: 'lr 2/17/2010 15:12'!
seasidePharoContinuation
	^ self new
		name: 'Seaside-Pharo-Continuation';
		description: 'Pharo implementation of full and partial first-class continuations.';
		addDependency: 'Grease-Pharo-Core';
		url: #seasideUrl;
		yourself! !

GRError subclass: #WAMarkerNotFound
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Pharo-Continuation'!

GRNotification subclass: #WAMarkerRequest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Pharo-Continuation'!

!WAMarkerRequest methodsFor: 'accessing' stamp: 'lr 2/5/2009 08:54'!
defaultAction
	^ nil! !

!ContextPart methodsFor: '*seaside-pharo-continuation' stamp: 'avi 4/24/2004 12:27'!
localAt: aNumber
	^ self at: aNumber! !

!ContextPart methodsFor: '*seaside-pharo-continuation' stamp: 'avi 4/24/2004 12:27'!
localAt: aNumber put: anObject
	^ self at: aNumber put: anObject! !

!ContextPart methodsFor: '*seaside-pharo-continuation' stamp: 'avi 4/24/2004 12:27'!
localSize
	^ self size! !

!ContextPart methodsFor: '*seaside-pharo-continuation' stamp: 'pmm 12/4/2007 20:31'!
serialiseOn: aStream
	aStream nextPut: self.
	1 to: self class instSize do: [ :index |
		aStream nextPut: (self instVarAt: index) ].
	1 to: self localSize do: [ :index |
		aStream nextPut: (self localAt: index) ].! !
